//  Copyright (c) 2001-2010 Hartmut Kaiser
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#if !defined(BOOST_SPIRIT_KARMA_OPERATOR_FEB_28_2007_0351PM)
#define BOOST_SPIRIT_KARMA_OPERATOR_FEB_28_2007_0351PM

#if defined(_MSC_VER)
#pragma once
#endif

#include <boost/spirit/home/karma/operator/sequence.hpp>
#include <boost/spirit/home/karma/operator/and_predicate.hpp>
#include <boost/spirit/home/karma/operator/not_predicate.hpp>
#include <boost/spirit/home/karma/operator/alternative.hpp>
#include <boost/spirit/home/karma/operator/kleene.hpp>
#include <boost/spirit/home/karma/operator/plus.hpp>
#include <boost/spirit/home/karma/operator/optional.hpp>
#include <boost/spirit/home/karma/operator/list.hpp>

#endif
