/*=============================================================================
    Copyright (c) 2001-2010 Joel de Guzman

    Distributed under the Boost Software License, Version 1.0. (See accompanying
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/
#if !defined(BOOST_SPIRIT_SUPPORT_SEPTEMBER_26_2008_0340AM)
#define BOOST_SPIRIT_SUPPORT_SEPTEMBER_26_2008_0340AM

#if defined(_MSC_VER)
#pragma once
#endif

#include<boost/spirit/home/support/assert_msg.hpp>
#include<boost/spirit/home/support/action_dispatch.hpp>
#include<boost/spirit/home/support/argument.hpp>
#include<boost/spirit/home/support/attributes.hpp>
#include<boost/spirit/home/support/char_class.hpp>
#include<boost/spirit/home/support/common_terminals.hpp>
#include<boost/spirit/home/support/container.hpp>
#include<boost/spirit/home/support/context.hpp>
#include<boost/spirit/home/support/info.hpp>
#include<boost/spirit/home/support/lazy.hpp>
#include<boost/spirit/home/support/make_component.hpp>
#include<boost/spirit/home/support/meta_compiler.hpp>
#include<boost/spirit/home/support/modify.hpp>
#include<boost/spirit/home/support/sequence_base_id.hpp>
#include<boost/spirit/home/support/string_traits.hpp>
#include<boost/spirit/home/support/terminal.hpp>
#include<boost/spirit/home/support/unused.hpp>
#include<boost/spirit/home/support/utf8.hpp>

#endif

