/*=============================================================================
    Copyright (c) 2004 Angus Leeming
    Copyright (c) 2006 Daniel Wallin
    Copyright (c) 2005 Dan Marsden
    Copyright (c) 2007 Joel de Guzman

    Distributed under the Boost Software License, Version 1.0. (See accompanying 
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/
#ifndef PHOENIX_STL_ALGORITHM_HPP
#define PHOENIX_STL_ALGORITHM_HPP

#include <boost/spirit/home/phoenix/stl/algorithm/iteration.hpp>
#include <boost/spirit/home/phoenix/stl/algorithm/querying.hpp>
#include <boost/spirit/home/phoenix/stl/algorithm/transformation.hpp>

#endif // PHOENIX_STL_ALGORITHM_HPP
