//  Copyright (c) 2001-2010 Hartmut Kaiser
//
//  Distributed under the Boost Software License, Version 1.0. (See accompanying
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#if !defined(BOOST_SPIRIT_KARMA_STREAM_MAY_01_2007_1254AM)
#define BOOST_SPIRIT_KARMA_STREAM_MAY_01_2007_1254AM

#if defined(_MSC_VER)
#pragma once
#endif

#include <boost/spirit/home/support/iterators/ostream_iterator.hpp>
#include <boost/spirit/home/karma/stream/stream.hpp>

#endif
