//  Copyright (c) 2001-2010 Hartmut Kaiser
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#if !defined(SPIRIT_KARMA_CORE_MARCH_06_2007_0833PM)
#define SPIRIT_KARMA_CORE_MARCH_06_2007_0833PM

#if defined(_MSC_VER)
#pragma once
#endif

#include <boost/spirit/home/karma/char.hpp>
#include <boost/spirit/home/karma/string.hpp>
#include <boost/spirit/home/karma/numeric.hpp>
#include <boost/spirit/home/karma/operator.hpp>
#include <boost/spirit/home/karma/nonterminal.hpp>
#include <boost/spirit/home/karma/action.hpp>
#include <boost/spirit/home/karma/directive.hpp>
#include <boost/spirit/home/karma/auxiliary.hpp>
#include <boost/spirit/home/karma/binary.hpp>
#include <boost/spirit/home/karma/generate.hpp>
#include <boost/spirit/home/karma/generate_attr.hpp>
#include <boost/spirit/home/karma/generator.hpp>
#include <boost/spirit/home/karma/delimit_out.hpp>
#include <boost/spirit/home/karma/what.hpp>
#include <boost/spirit/home/karma/stream.hpp>
#include <boost/spirit/home/karma/auto.hpp>
#include <boost/spirit/home/karma/format.hpp>

#endif
