/*=============================================================================
    Copyright (c) 2007 Joel de Guzman

    Distributed under the Boost Software License, Version 1.0. (See accompanying
    file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/
#ifndef PHOENIX_FUSION_HPP
#define PHOENIX_FUSION_HPP

#include <boost/spirit/home/phoenix/fusion/at.hpp>

#endif
